import numpy as np
import networkx as nx

EARTH_RADIUS_M = 6_371_009  # distance in meters


def haversine_dist(lat1, lon1, lat2, lon2, earth_radius=EARTH_RADIUS_M):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(np.deg2rad, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    # haversine formula 
    h = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    h = np.minimum(1, h)  # protect against floating point errors
    arc = 2 * np.arcsin(np.sqrt(h))

    # return distance in units of earth_radius
    return arc * earth_radius


def enlarge_bbox(north, south, west, east, dist):
    delta_lat = (dist / EARTH_RADIUS_M) * (180 / np.pi)
    lat_mean = np.mean([north, south])
    delta_lng = (dist / EARTH_RADIUS_M) * (180 / np.pi) / np.cos(lat_mean * np.pi / 180)
    north = north + delta_lat
    south = south - delta_lat
    east = east + delta_lng
    west = west - delta_lng

    return north, south, west, east


def add_edge_lengths(G, precision=3):
    uvk = tuple(G.edges)

    lat = G.nodes(data='y')
    lon = G.nodes(data='x')

    lat_u, lon_u, lat_v, lon_v = list(zip(*[(lat[u], lon[u], lat[v], lon[v]) for u, v, _ in uvk]))

    dists = haversine_dist(lat_u, lon_u, lat_v, lon_v).round(precision)
    dists[np.isnan(dists)] = 0
    nx.set_edge_attributes(G, values=dict(zip(uvk, dists)), name='length')
    return G



import json
import os
from pathlib import Path

import requests


def extract_streetview_pic(point, api_key, size="640x640", heading=90, pitch=-10):
    meta_base = 'https://maps.googleapis.com/maps/api/streetview/metadata?'
    pic_base = 'https://maps.googleapis.com/maps/api/streetview?'

    lat = point[0]
    lon = point[1]

    meta_params = {'key': api_key,
                   'location': f'{lat}, {lon}',
                   'source': 'outdoor'}

    pic_params = {
        'size': size,  # max 640x640 pixels
        'location': f'{lat}, {lon}',
        'heading': str(heading),
        'pitch': str(pitch),
        'key': api_key,
        'source': 'outdoor'
    }

    # Do the request and get the response data
    meta_response = requests.get(meta_base, params=meta_params)
    meta = meta_response.json()
    meta_response.close()
    # if (meta["pano_id"] != PREV_PAN_ID) and (meta["status"] == "OK"):
    #    PREV_PAN_ID = meta["pano_id"]
    #    pic_response = requests.get(pic_base, params=pic_params)

    #       pic = pic_response.content
    #        pic_response.close()
    #  else:
    #     meta = None
    #    pic = None
    pic_response = requests.get(pic_base, params=pic_params)
    pic = pic_response.content
    pic_response.close()

    return pic, meta


def save_streetview(pic, meta, folder_path):
    Path(os.path.join(folder_path)).mkdir(parents=True, exist_ok=True)

    with open(os.path.join(folder_path, 'pic.png'), 'wb') as file:
        file.write(pic)

    with open(os.path.join(folder_path, 'metadata.json'), 'w+') as out_file:
        json.dump(meta, out_file)



import datetime as dt
from itertools import groupby

import networkx as nx
from shapely.geometry import Point, LineString

from . import distance
from . import download

useful_tags_way = [
    "bridge",
    "tunnel",
    "oneway",
    "lanes",
    "ref",
    "name",
    "highway",
    "maxspeed",
    "service",
    "access",
    "area",
    "landuse",
    "width",
    "est_width",
    "junction",
]


def graph_from_bbox(north, south, west, east, simplify=True, network_type='drive', buffer_dist=0):
    bbox_buffer = distance.enlarge_bbox(north, south, west, east, buffer_dist)
    response_json = download.osm_download(bbox_buffer, network_type=network_type)
    G = create_graph(response_json)

    if simplify:
        G.graph["simplified"] = True
        G = _simplification(G, response_json)
    else:
        G.graph["simplified"] = False
    return G


def _to_simplify(segment):
    if len(segment) > 2:
        return True
    else:
        return False


def _simplification(G, response_json):
    _, paths = get_nodes_edges(response_json)
    for path in paths.values():
        is_oneway = _is_oneway(path, False)

        nodes = path.pop("nodes")

        junction = [False, False]
        if is_oneway:
            junction[1:1] = [True if G.degree[node] > 3 else False for node in nodes[1:-1]]
        else:
            junction[1:1] = [True if G.degree[node] > 4 else False for node in nodes[1:-1]]

        all_nodes_to_remove = []
        all_edges_to_add = []

        for segment in _split_at_values(nodes, junction):
            if _to_simplify(segment):
                all_edges_to_add.append(_build_edge(G, segment, path, is_oneway, reverse=False))

                if not is_oneway:
                    all_edges_to_add.append(_build_edge(G, segment, path, is_oneway, reverse=True))

                all_nodes_to_remove.extend(segment[1:-1])

        G.remove_nodes_from(set(all_nodes_to_remove))

        for edge in all_edges_to_add:
            G.add_edge(edge["origin"], edge["destination"], **edge["attr_dict"])

    return G


def _build_edge_attribute(G, segment, segment_attributes, is_oneway):
    attribute = segment_attributes.copy()
    #attribute = {k: segment_attributes[k] for k in useful_tags_way if k in segment_attributes}

    attribute["geometry"] = LineString([Point((G.nodes[node]["x"],
                                               G.nodes[node]["y"])) for node in segment])
    attribute["oneway"] = is_oneway

    lat, lon = zip(*[(G.nodes[node]["y"], G.nodes[node]["x"]) for node in segment])
    edge_lengths = sum(distance.haversine_dist(lat[:-1], lon[:-1], lat[1:], lon[1:])).round(3)

    attribute["length"] = edge_lengths

    return attribute


def _build_edge(G, segment, segment_attributes, is_oneway, reverse=False):
    if not reverse:
        segment_attributes = _build_edge_attribute(G, segment, segment_attributes, is_oneway)

        edge = {"origin": segment[0], "destination": segment[-1],
                "attr_dict": segment_attributes}
        return edge
    else:
        segment_attributes = _build_edge_attribute(G, segment[::-1], segment_attributes, is_oneway)
        edge = {"origin": segment[-1], "destination": segment[0],
                "attr_dict": segment_attributes}
        return edge


def _split_at_values(nodes, junction):
    indices = [i for i, x in enumerate(nodes) if junction[i]]
    for start, end in zip([0, *indices], [*indices, len(nodes)]):
        yield nodes[start:end + 1]


def create_graph(response_json):
    # TODO: nella creazione del grafo rimuovi intersezioni multiple, in modo tale che c'è un solo nodo intersezione
    #  per grafo create the graph as a MultiDiGraph and set its meta-attributes
    metadata = {
        'created_date': "{:%Y-%m-%d %H:%M:%S}".format(dt.datetime.now()),
        'created_with': f"PyTrack 1.0.0",
        'crs': "epsg:4326",
    }
    G = nx.MultiDiGraph(**metadata)

    nodes, paths = get_nodes_edges(response_json)

    # add each osm node to the graph
    for node, data in nodes.items():
        G.add_node(node, **data)

    add_edges(G, paths, bidirectional=False)

    # add length (haversine distance between nodes) attribute to each edge
    if len(G.edges) > 0:
        G = distance.add_edge_lengths(G)

    return G


def _oneway_path_values(path):
    return {path[key] for key in path.keys() if key.startswith("oneway") and path[key] == "no"}


def _is_oneway(path, bidirectional):
    no_oneway_values = {"no", "false", "0", "reversible", "alternating"}

    oneway_path_values = _oneway_path_values(path)
    is_oneway = oneway_path_values.isdisjoint(no_oneway_values) and (
        not not oneway_path_values) and not bidirectional

    return is_oneway


def _is_reversed(path):
    reversed_values = {"-1", "reverse", "T"}

    oneway_path_values = _oneway_path_values(path)
    is_reversed = not oneway_path_values.isdisjoint(reversed_values)

    return is_reversed


def add_edges(G, paths, bidirectional=False, all_oneway=False):
    # TODO: "junction": "roundabout" è oneway
    for path in paths.values():
        nodes = path.pop('nodes')

        is_oneway = _is_oneway(path, bidirectional)
        is_reversed = _is_reversed(path)

        if is_oneway and is_reversed:
            nodes.reverse()

        if not all_oneway:
            path['oneway'] = is_oneway

        edges = list(zip(nodes[:-1], nodes[1:]))
        if not is_oneway:
            edges.extend([(v, u) for u, v in edges])

        G.add_edges_from(edges, **path)


def convert_edge(element):
    # add OSM path id and remove consecutive duplicate nodes in the list of path's nodes
    path = {'osmid': element['id'], 'nodes': [group for group, _ in groupby(element['nodes'])]}

    # add tags
    path.update(element['tags'])

    return path


def convert_node(element):
    # add node's GPS coordinates
    node = {'y': element['lat'], 'x': element['lon']}

    # add tags
    node.update(element['tags']) if "tags" in element else None

    return node


def get_nodes_edges(response_json):
    nodes = dict()
    paths = dict()

    for element in response_json['elements']:
        if element['type'] == "node":
            nodes[element['id']] = convert_node(element)
        elif element['type'] == "way":
            paths[element['id']] = convert_edge(element)

    return nodes, paths

# %%
# grafo = graph_from_bbox(41.86722, 41.86660, 12.41930, 12.42239, simplify=True, network_type='drive')



from collections import deque

from . import mpmatching_utils


def viterbi_search(G, trellis, start, target):
    # Initialize joint probability for each node
    joint_prob = {}
    for u_name in trellis.nodes():
        joint_prob[u_name] = 0
    predecessor = {}
    queue = deque()

    queue.append(start)
    joint_prob[start] = mpmatching_utils.emission_prob(trellis.nodes[start]["candidate"])
    predecessor[start] = None

    while queue:
        # Extract node u
        u_name = queue.popleft()
        u = trellis.nodes[u_name]["candidate"]

        if u_name == target:
            break
        for v_name in trellis.successors(u_name):
            v = trellis.nodes[v_name]["candidate"]

            new_prob = joint_prob[u_name] * mpmatching_utils.transition_prob(G, u, v) * mpmatching_utils.emission_prob(
                v)
            if joint_prob[v_name] < new_prob:
                joint_prob[v_name] = new_prob
                predecessor[v_name.split("_")[0]] = u_name
            if v_name not in queue:
                queue.append(v_name)

    return joint_prob[target], predecessor



import itertools
from decimal import Decimal

import folium
import networkx as nx
import numpy as np
import pandas as pd
from pyproj import Geod
from shapely.geometry import Point, LineString
from sklearn.neighbors import BallTree

from pytrack.graph import distance, plot, utils

geod = Geod(ellps="WGS84")


def get_unique_number(lon, lat):
    if isinstance(lat, str):
        lat_double = float(lat)
    else:
        lat_double = lat
    if isinstance(lon, str):
        lon_double = float(lon)
    else:
        lon_double = lon

    lat_int = int((lat_double * 10 ** abs(Decimal(str(lat_double)).as_tuple().exponent)))
    lon_int = int((lon_double * 10 ** abs(Decimal(str(lon_double)).as_tuple().exponent)))

    val = abs((lat_int << 16 & 0xffff0000) | (lon_int & 0x0000ffff))
    val = val % 2147483647

    # alternative use hash or other
    # return int(str(lat_int)+str(lon_int))
    return val


def interpolate_graph(G, dist=1):
    G = G.copy()

    edges_toadd = []
    nodes_toadd = []

    for edge in list(G.edges(data=True)):
        u, v, data = edge
        # oneway = data["oneway"]
        geom = data["geometry"]
        data["length"] = dist

        G.remove_edge(u, v)

        XY = [xy for xy in _interpolate_geom(geom, dist=dist)]

        # generate nodes id
        uv = [(get_unique_number(*u), get_unique_number(*v)) for u, v in zip(XY[:-1], XY[1:])]
        # edges_interp = [LineString([u, v]) for u, v in zip(XY[:-1], XY[1:])]
        data_edges = [(lambda d: d.update({"geometry": LineString([u, v])}) or d)(data.copy()) for u, v in
                      zip(XY[:-1], XY[1:])]
        edges_toadd.extend([(*uv, data) for uv, data in zip(uv, data_edges)])

        nodes_toadd.extend([(get_unique_number(lon, lat),
                             {"x": lon, "y": lat, "geometry": Point(lon, lat)}) for lon, lat in XY])

        """
        # if not oneway:
        #  G.remove_edge(v, u)
        # uv_inv = [(get_unique_number(*u), get_unique_number(*v)) for u, v in zip(XY[:-len(XY):-1], XY[-2::-1])]
        #  uv_inv = [(v, u) for u, v in uv[::-1]]
        # edges_interp_inv = [LineString([u, v]) for u, v in zip(XY[:-len(XY):-1], XY[-2::-1])] 
        #  data_inv_edges = [(lambda d: d.update({"geometry": LineString([u, v])}) or d)(data.copy()) \
        #  for u, v in zip(XY[:-len(XY):-1], XY[-2::-1])]
        #  edges_toadd.extend([(*uv, data) for uv, data in zip(uv_inv, data_inv_edges)])
        """
    G.add_edges_from(edges_toadd)
    G.add_nodes_from(nodes_toadd)

    G.remove_nodes_from(list(nx.isolates(G)))
    G.interpolation = True
    return G


def _interpolate_geom(geom, dist=1):
    # TODO: use geospatial interpolation see: npts method in https://pyproj4.github.io/pyproj/stable/api/geod.html
    num_vert = max(round(geod.geometry_length(geom) / dist), 1)
    for n in range(num_vert + 1):
        point = geom.interpolate(n / num_vert, normalized=True)
        yield point.x, point.y


def interpolate_geom(geom, dist=1):
    if isinstance(geom, LineString):
        return LineString([xy for xy in _interpolate_geom(geom, dist)])


def get_candidates(G, points, interp_dist=1, closest=True, radius=10):
    """
    G: graph
    x: longitude
    y: latitude
    distance: distance of interpolation
    radius: radius of research
    """

    G = G.copy()

    if interp_dist:
        G = interpolate_graph(G, dist=interp_dist)

    geoms = utils.graph_to_gdfs(G, nodes=False).set_index(["u", "v"])[["osmid", "geometry"]]

    uv_xy = [[u, osmid, np.deg2rad(xy)] for uv, geom, osmid in
             zip(geoms.index, geoms.geometry.values, geoms.osmid.values)
             for u, xy in zip(uv, geom.coords[:])]

    index, osmid, xy = zip(*uv_xy)

    nodes = pd.DataFrame(xy, index=osmid, columns=["x", "y"])[["y", "x"]]

    ball = BallTree(nodes, metric='haversine')

    idxs, dists = ball.query_radius(np.deg2rad(points), radius / distance.EARTH_RADIUS_M, return_distance=True)
    dists = dists * distance.EARTH_RADIUS_M  # radians to meters

    if closest:
        results = dict()
        for i, (point, idx, dist) in enumerate(zip(points, idxs, dists)):
            df = pd.DataFrame({"osmid": list(np.array(index)[idx]),
                               "edge_osmid": list(nodes.index[idx]),
                               "coords": tuple(map(tuple, np.rad2deg(nodes.values[idx]))),
                               "dist": dist})

            df = df.loc[df.groupby('edge_osmid')['dist'].idxmin()].reset_index(drop=True)

            results[i] = {"observation": point,
                          "osmid": list(df["osmid"]),
                          "edge_osmid": list(df["edge_osmid"]),
                          "candidates": list(df["coords"]),
                          "candidate_type": np.full(len(df["coords"]), False),
                          "dists": list(df["dist"])}

    else:
        results = {i: {"observation": point,
                       "osmid": list(np.array(index)[idx]),
                       "edge_osmid": list(nodes.index[idx]),
                       "candidates": list(map(tuple, np.rad2deg(nodes.values[idx]))),
                       "candidate_type": np.full(len(nodes.index[idx]), False),
                       "dists": list(dist)} for i, (point, idx, dist) in enumerate(zip(points, idxs, dists))}

    return G, results


def get_candidates_old(G, points, interp_dist=1, closest=True, radius=10):
    """
    G: graph
    x: longitude
    y: latitude
    distance: distance of interpolation
    radius: radius of research
    """
    geoms = utils.graph_to_gdfs(G, nodes=False).set_index(["u", "v", "key"])[["osmid", "geometry"]]

    uvk_xy = [[uvk, osmid, np.deg2rad(xy)] for uvk, geom, osmid in
              zip(geoms.index, geoms.geometry.values, geoms.osmid.values)
              for xy in _interpolate_geom(geom, dist=interp_dist)]
    index, osmid, xy = zip(*uvk_xy)

    nodes = pd.DataFrame(xy, index=osmid, columns=["x", "y"])[["y", "x"]]

    ball = BallTree(nodes, metric='haversine')

    idxs, dists = ball.query_radius(np.deg2rad(points), radius / distance.EARTH_RADIUS_M, return_distance=True)
    dists = dists * distance.EARTH_RADIUS_M  # radians to meters

    if closest:
        results = dict()
        for i, (point, idx, dist) in enumerate(zip(points, idxs, dists)):
            df = pd.DataFrame({"edge_osmid": list(nodes.index[idx]),
                               "coords": tuple(map(tuple, np.rad2deg(nodes.values[idx]))),
                               "dist": dist})

            df = df.loc[df.groupby('edge_osmid')['dist'].idxmin()].reset_index(drop=True)

            results[i] = {"observation": point,
                          "edge_osmid": list(df["edge_osmid"]),
                          "candidates": list(df["coords"]),
                          "dists": list(df["dist"])}

    else:
        results = {i: {"observation": point,
                       "edge_osmid": list(nodes.index[idx]),
                       "candidates": list(map(tuple, np.rad2deg(nodes.values[idx]))),
                       "dists": list(dist)} for i, (point, idx, dist) in enumerate(zip(points, idxs, dists))}
    return results


def plot_candidates(G, candidates, radius):
    graph_map = plot.plot_graph_folium(G)

    for i, obs in enumerate(candidates.keys()):
        folium.Circle(location=candidates[obs]["observation"], radius=radius, weight=1, color="black", fill=True,
                      fill_opacity=0.2).add_to(graph_map)
        popup = f'{i}-th point \n Latitude: {candidates[obs]["observation"][0]}\n Longitude: ' \
                f'{candidates[obs]["observation"][1]}'
        folium.Circle(location=candidates[obs]["observation"], popup=popup, radius=1, color="black",
                      fill=True, fill_opacity=1).add_to(graph_map)

        # plot candidates
        for cand, label, cand_type in zip(candidates[obs]["candidates"], candidates[obs]["edge_osmid"],
                                          candidates[obs]["candidate_type"]):
            popup = f"coord: {cand} \n edge_osmid: {label}"
            if cand_type:
                folium.Circle(location=cand, popup=popup, radius=2, color="yellow", fill=True, fill_opacity=1).add_to(
                    graph_map)
            else:
                folium.Circle(location=cand, popup=popup, radius=1, color="red", fill=True, fill_opacity=1).add_to(
                    graph_map)

    folium.LatLngPopup().add_to(graph_map)

    return graph_map


def elab_candidate_results(results, predecessor):
    results = results.copy()
    for key in list(results.keys())[:-1]:
        win_cand_idx = int(predecessor[str(key + 1)].split("_")[1])
        results[key]["candidate_type"][win_cand_idx] = True

    win_cand_idx = int(predecessor["target"].split("_")[1])
    results[list(results.keys())[-1]]["candidate_type"][win_cand_idx] = True
    return results


def create_path(G, trellis, predecessor):
    u, v = list(zip(*[(u, v) for v, u in predecessor.items()][2:]))
    path = [(u, v) for u, v in zip(u, u[1:])]

    path_elab = [node for u, v in path for node in nx.shortest_path(G, trellis.nodes[u]["candidate"].node_id,
                                                                    trellis.nodes[v]["candidate"].node_id,
                                                                    weight='length')]
    path_elab = [k for k, g in itertools.groupby(path_elab)]
    return path_elab


def plot_path(G: object, results: object, trellis: object, predecessor: object, radius: object) -> object:
    # TODO: non ricreare altra mappa ma utilizza quella esistente oppure utinizzane una semplificata
    path_map = plot_candidates(G, results, radius)

    path_elab = create_path(G, trellis, predecessor)

    edge_attr = dict()
    edge_attr["color"] = "green"
    edge_attr["weight"] = 4
    edge_attr["opacity"] = 1

    edge = [(lat, lng) for lng, lat in LineString([G.nodes[node]["geometry"] for node in path_elab]).coords]
    folium.PolyLine(locations=edge, **edge_attr, ).add_to(path_map)

    return path_map



from shapely.geometry import LineString, Point
import pandas as pd

'''
ISSUE: geopandas ha problemi con osx el captain
def graph_to_gdfs(G, nodes=True, edges=True, node_geometry=True, edge_geometry=True):
    #TODO: gestisci il fatto che grafo sia semplificato e quindi geometrie esistono
    crs = G.graph["crs"]

    if nodes:
        nodes, data = zip(*G.nodes(data=True))

        if node_geometry:
            # convert node x/y attributes to Points for geometry column
            geom = [Point(d["x"], d["y"]) for d in data]
            gdf_nodes = gpd.GeoDataFrame(data, crs=crs, geometry=geom)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)
        else:
            gdf_nodes = gpd.GeoDataFrame(data)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)

    if edges:
        u, v, k, data = zip(*G.edges(keys=True, data=True))
        if edge_geometry:
            longs = G.nodes(data="x")
            lats = G.nodes(data="y")

            if G.graph["simplified"]:
                return
            else:
                geoms = [LineString((Point((longs[src], lats[src])),
                                     Point((longs[tgt], lats[tgt])))) for src, tgt in zip(u, v)]
            
            gdf_edges = gpd.GeoDataFrame(data, crs=crs, geometry=geoms)

        else:
            gdf_edges = gpd.GeoDataFrame(data)
            gdf_edges["geometry"] = None
            gdf_edges.crs = crs

        gdf_edges["u"], gdf_edges["v"], gdf_edges["key"] = u, v, k

        gdf_edges = gdf_edges[["u", "v", "key"] + gdf_edges.columns.to_list()[:-3]]

    if nodes and edges:
        return gdf_nodes, gdf_edges
    elif nodes:
        return gdf_nodes
    elif edges:
        return gdf_edges
'''


def graph_to_gdfs(G, nodes=True, edges=True, node_geometry=True, edge_geometry=True):
    crs = G.graph["crs"]

    if nodes:
        nodes, data = zip(*G.nodes(data=True))

        if node_geometry:
            # convert node x/y attributes to Points for geometry column
            for d in data:
                d["geometry"] = Point(d["x"], d["y"])
            gdf_nodes = pd.DataFrame(data)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)
        else:
            gdf_nodes = pd.DataFrame(data)
            gdf_nodes.insert(loc=0, column='osmid', value=nodes)

    if edges:
        u, v, k, data = zip(*G.edges(keys=True, data=True))
        if edge_geometry:
            longs = G.nodes(data="x")
            lats = G.nodes(data="y")

            for d, src, tgt in zip(data, u, v):
                if "geometry" not in d:
                    d["geometry"] = LineString((Point((longs[src], lats[src])),
                                                Point((longs[tgt], lats[tgt]))))
            gdf_edges = pd.DataFrame(data)

        else:
            gdf_edges = pd.DataFrame(data)
            gdf_edges["geometry"] = None
            gdf_edges.crs = crs

        gdf_edges["u"], gdf_edges["v"], gdf_edges["key"] = u, v, k

        gdf_edges = gdf_edges[["u", "v", "key"] + gdf_edges.columns.to_list()[:-3]]

    if nodes and edges:
        return gdf_nodes, gdf_edges
    elif nodes:
        return gdf_nodes
    elif edges:
        return gdf_edges



import folium

from . import utils


def plot_graph_folium(G, plot_nodes=False, edge_color="#3388ff", edge_width=3,
                      edge_opacity=1, radius=1.7, node_color="red", fill=True, fill_color=None,
                      fill_opacity=1):
    edge_attr = dict()
    edge_attr["color"] = edge_color
    edge_attr["weight"] = edge_width
    edge_attr["opacity"] = edge_opacity

    node_attr = dict()
    node_attr["color"] = node_color
    node_attr["fill"] = fill
    if not fill_color:
        node_attr["fill_color"] = node_color
    else:
        node_attr["fill_color"] = node_color
    node_attr["fill_opacity"] = fill_opacity

    nodes, edges = utils.graph_to_gdfs(G)
    edges = utils.graph_to_gdfs(G, nodes=False)

    graph_map = folium.Map([nodes.y.mean(), nodes.x.mean()], tiles='CartoDB positron', zoom_start=16)
    folium.LatLngPopup().add_to(graph_map)

    for geom in edges.geometry:
        edge = [(lat, lng) for lng, lat in geom.coords]
        folium.PolyLine(locations=edge, **edge_attr, ).add_to(graph_map)

    if plot_nodes:
        for point, osmid in zip(nodes.geometry, nodes.osmid):
            folium.Circle(location=(point.y, point.x), popup=f"osmid: {osmid}", radius=radius, **node_attr).add_to(
                graph_map)

    return graph_map



import math

import matplotlib.pyplot as plt
import networkx as nx
from pytrack.graph import distance

SIGMA_Z = 4.07
BETA = 20


# A gaussian distribution
def emission_prob(u):
    if u.great_dist:
        c = 1 / (SIGMA_Z * math.sqrt(2 * math.pi))
        return c * math.exp(-(u.great_dist / SIGMA_Z) ** 2)
    else:
        return 1


# A empirical distribution
def transition_prob(G, u, v):   
    c = 1 / BETA

    if u.great_dist and v.great_dist:
        delta = abs(
            nx.shortest_path_length(G, u.node_id, v.node_id, weight="length", method='dijkstra')
            - distance.haversine_dist(*u.coord, *v.coord))
        return c * math.exp(-delta / BETA) * 1e5
    else:
        return 1


class Candidate:
    __slots__ = ["node_id", "edge_osmid", "obs", "great_dist", "coord"]

    def __init__(self, node_id, edge_osmid, obs, great_dist, coord):
        self.node_id = node_id
        self.edge_osmid = edge_osmid
        self.obs = obs
        self.great_dist = great_dist
        self.coord = coord


def create_trellis(results):
    G = nx.DiGraph()

    prev_nodes = ["start"]
    G.add_node("start", candidate=Candidate("start", None, None, None, None))
    G.add_node("target", candidate=Candidate("start", None, None, None, None))

    for idx, item in results.items():
        obs = item["observation"]
        nodes, data_node = zip(*[(f"{idx}_{j}", {"candidate": Candidate(node_id, edge_osmid, obs, dist, cand)})
                                 for j, (node_id, edge_osmid, dist, cand) in
                                 enumerate(zip(item["osmid"], item["edge_osmid"], item["dists"], item["candidates"]))])
        edges = [(u, v) for u in prev_nodes for v in nodes]

        G.add_nodes_from(zip(nodes, data_node))
        G.add_edges_from(edges)
        prev_nodes = nodes

    edges = [(u, "target") for u in prev_nodes]
    G.add_edges_from(edges)
    return G


def draw_trellis(T, figsize=(15, 12), dpi=300, node_size=500, font_size=8):
    plt.figure(1, figsize=figsize, dpi=dpi)

    pos = nx.drawing.nx_pydot.graphviz_layout(T, prog='dot', root='start')
    trellis_diag = nx.draw_networkx(T, pos, node_size=node_size, font_size=font_size)

    return trellis_diag



import requests


def get_filters(network_type='drive'):
    osm_filters = dict()

    osm_filters['drive'] = ('["highway"]["area"!~"yes"]["access"!~"private"]'
                            '["highway"!~"abandoned|bridleway|bus_guideway|construction|corridor|cycleway|'
                            'elevator|escalator|footway|path|pedestrian|planned|platform|proposed|raceway|steps|track"]'
                            '["service"!~"emergency_access|parking|parking_aisle|private"]')

    return osm_filters[network_type]


def osm_download(bbox, network_type=None):
    # TODO: add network_type statement
    north, south, west, east = bbox

    osm_filters = get_filters(network_type='drive')

    url_endpoint = 'https://maps.mail.ru/osm/tools/overpass/api/interpreter'

    timeout = 180
    out_resp = 'json'

    overpass_settings = f'[out:{out_resp}][timeout:{timeout}]'

    query_str = f'{overpass_settings};(way{osm_filters}({south}, {west}, {north}, {east});>;);out;'

    response_json = requests.post(url_endpoint, data={'data': query_str})

    size_kb = len(response_json.content) / 1000
    print(f'Downloaded {size_kb:,.2f}kB')

    return response_json.json()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='PyTrack-lib',
    version='1.0.0',
    package_dir={"": "pytrack"},
    packages=setuptools.find_packages(where="pytrack"),
    url='https://github.com/cosbidev/PyTrack',
    license='BSD-3-Clause-Clear',
    author='Matteo Tortora',
    author_email='m.tortora@unicampus.it',
    description='a Map-Matching-based Python Toolbox for Vehicle Trajectory Reconstruction',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ]
)
